﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web.Mvc;

namespace Harold.Net.Web.Mvc.Extension
{
    public class ConfiguredControllerActionInvoker : ControllerActionInvoker
    {
        protected override IDictionary<string, object> GetParameterValues(ControllerContext controllerContext, ActionDescriptor actionDescriptor)
        {
            // controllerContext.Controller.ViewData.ModelState

            var obj = base.GetParameterValues(controllerContext, actionDescriptor);
            return obj;
        }

        protected override object GetParameterValue(ControllerContext controllerContext, ParameterDescriptor parameterDescriptor)
        {
            object rawValue = base.GetParameterValue(controllerContext, parameterDescriptor);

            object newValue = RefineParameterValue(rawValue, parameterDescriptor.ParameterType);
            ValidateParameter(controllerContext, parameterDescriptor, newValue);

            return newValue;
        }

        protected virtual object RefineParameterValue(object rawValue, Type parameterType)
        {
            object newValue = rawValue;

            if (rawValue == null && parameterType.IsValueType)
            {
                try
                {
                    newValue = Activator.CreateInstance(parameterType);
                }
                catch (Exception e)
                {
                    // should handle the exception.
                }
            }

            return newValue;
        }

        protected virtual void ValidateParameter(ControllerContext controllerContext, ParameterDescriptor parameterDescriptor, object parameterValue)
        {
            Dictionary<string, bool> startedValid = new Dictionary<string, bool>(StringComparer.OrdinalIgnoreCase);

            Dictionary<ModelMetadata, IEnumerable<ModelValidator>> validators = new Dictionary<ModelMetadata,IEnumerable<ModelValidator>>();
            var metadata = ModelMetadataProviders.Current.GetMetadataForType(() => parameterValue, null, parameterDescriptor.ParameterType, parameterDescriptor.ParameterName);
            GetValidators(controllerContext, metadata, validators);
            foreach (KeyValuePair<ModelMetadata, IEnumerable<ModelValidator>> kvp in validators)
            {
                string modelKey = kvp.Key.GetFullPropertyName();
                if (controllerContext.Controller.ViewData.ModelState.ContainsKey(modelKey))
                {
                    controllerContext.Controller.ViewData.ModelState[modelKey].Errors.Clear();
                }                

                foreach (ModelValidator validator in kvp.Value)
                {
                    startedValid[modelKey] = true;
                    foreach (ModelValidationResult result in validator.Validate(null))
                    {
                        startedValid[modelKey] = false;
                        controllerContext.Controller.ViewData.ModelState.AddModelError(modelKey, result.Message);
                        break;
                    }

                    if (!startedValid[modelKey])
                    {
                        break;
                    }
                }
               
            }
        }

        protected virtual void GetValidators(ControllerContext controllerContext, ModelMetadata modelMetadata, Dictionary<ModelMetadata, IEnumerable<ModelValidator>> validators)
        {
            if (!modelMetadata.IsComplexType)
            {
                validators.Add(modelMetadata, modelMetadata.GetValidators(controllerContext));
                return;
            }

            foreach (ModelMetadata metadata in modelMetadata.Properties)
            {
                GetValidators(controllerContext, metadata, validators);
            }
        }

        protected static bool IsModelValid(ControllerContext controllerContext, string modelKey)
        {
            return controllerContext.Controller.ViewData.ModelState.IsValidField(modelKey);
        }

        private static string CreateSubPropertyName(string prefix, string propertyName)
        {
            if (String.IsNullOrEmpty(prefix))
            {
                return propertyName;
            }
            else if (String.IsNullOrEmpty(propertyName))
            {
                return prefix;
            }
            else
            {
                return prefix + "." + propertyName;
            }
        }
    }
}
