﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Web.Mvc;
using System.Web.Routing;


using Microsoft.Practices.EnterpriseLibrary.Validation;
using Microsoft.Practices.EnterpriseLibrary.Validation.Configuration;

namespace Harold.Net.Web.Mvc.Extension
{
    public class EntlibValidationAdapter
    {
        Stack _hierarchy;
        static readonly ValidatorConfigCache _cache = new ValidatorConfigCache();

        ControllerContext _context;
        ControllerContext Context
        {
            get
            {
                return _context;
            }
        }

        public EntlibValidationAdapter(ControllerContext controllerContext)
        {
            _context = controllerContext;
            
        }

        public virtual IEnumerable<ModelValidator> GetModelValidators(ModelMetadata modelMetadata)
        {
            var modelConfig = GetModelConfig(modelMetadata);
            return GenerateValidators(modelConfig, modelMetadata);
        }

        protected virtual ModelSection GetModelConfig(ModelMetadata modelMetadata)
        {
            ModelSection modelConfig = null;

            if (modelMetadata != null)
            {
                if (_cache.ContainKey(modelMetadata.GetFullPropertyName()))
                {
                    modelConfig = _cache.GetConfig(modelMetadata.GetFullPropertyName());
                }
                else
                {
                    _hierarchy = new System.Collections.Stack();
                    ConstructHierarchy(modelMetadata);
                    ModelMetadata rootMetadata = ((ModelMetadata)_hierarchy.Pop());
                    var rootModelConfig = GetRootModelConfig(GetController(Context.RequestContext), GetAction(Context.RequestContext), rootMetadata.PropertyName);
                    modelConfig = GetModelConfig(rootModelConfig, rootMetadata);
                    _cache.SaveConfig(modelMetadata.GetFullPropertyName(), modelConfig);
                }
            }

            return modelConfig;
        }

        protected virtual void ConstructHierarchy(ModelMetadata modelMetadata)
        {
            _hierarchy.Push(modelMetadata);

            ModelMetadata parentMetadata = modelMetadata.GetParentModelMetadata();
            if (parentMetadata != null)
            {
                ConstructHierarchy(parentMetadata);
            }
        }

        static ModelSection GetRootModelConfig(string controller, string action, string param)
        {
            if (string.IsNullOrEmpty(controller) || string.IsNullOrEmpty(action) || string.IsNullOrEmpty(param))
            {
                return null;
            }

            var validationConfig = ExtensionConfigManager.GetValidation();
            if (validationConfig == null)
            {
                return null;
            }

            var controllerConfig = validationConfig.Controllers[controller];
            if (controllerConfig == null)
            {
                return null;
            }

            var actionConfig = controllerConfig.Actions[action];
            if (actionConfig == null)
            {
                return null;
            }

            return GetRefModelConfig(actionConfig.Params[param]);
        }

        protected virtual ModelSection GetModelConfig(ModelSection modelConfig, ModelMetadata modelMetadata)
        {
            if (modelConfig == null)
            {
                return modelConfig;
            }
            modelConfig = GetRefModelConfig(modelConfig);

            if (_hierarchy.Count == 0)
            {
                return modelConfig;
            }

            modelMetadata = (ModelMetadata)_hierarchy.Pop();
            modelConfig = modelConfig.PropertyArray[modelMetadata.PropertyName];

            return GetModelConfig(modelConfig, modelMetadata);
        }

        protected virtual IEnumerable<ModelValidator> GenerateValidators(ModelSection modelConfig, ModelMetadata modelMetadata)
        {
            List<ModelValidator> validators = new List<ModelValidator>();

            if (modelConfig != null && modelMetadata != null && modelConfig.Validators != null)
            {
                foreach (ValidatorData vd in modelConfig.Validators)
                {
                    validators.Add(new EntlibModelValidator(modelMetadata, Context, ((IValidatorDescriptor)vd).CreateValidator(modelMetadata.ModelType, null, null, null)));
                }
            }

            return validators;
        }

        static ModelSection GetRefModelConfig(ModelSection modelConfig)
        {
            if (modelConfig == null || string.IsNullOrEmpty(modelConfig.Reference))
            {
                return modelConfig;
            }

            var validationConfig = ExtensionConfigManager.GetValidation();
            if (validationConfig == null)
            {
                return null;
            }

            return GetRefModelConfig(validationConfig.Models[modelConfig.Reference]);
        }

        static string GetController(RequestContext requestContext)
        {
            return string.Format("{0}Controller", requestContext.RouteData.GetRequiredString("controller"));
        }

        static string GetAction(RequestContext requestContext)
        {
            return requestContext.RouteData.GetRequiredString("action");
        }

        private static string CreateSubPropertyName(string prefix, string propertyName)
        {
            if (String.IsNullOrEmpty(prefix))
            {
                return propertyName;
            }
            else if (String.IsNullOrEmpty(propertyName))
            {
                return prefix;
            }
            else
            {
                return prefix + "." + propertyName;
            }
        }
    }

    public sealed class ValidatorConfigCache
    {
        readonly Dictionary<string, ModelSection> _cache = new Dictionary<string, ModelSection>();
        readonly ReaderWriterLockSlim _rwLock = new ReaderWriterLockSlim();

        public ValidatorConfigCache()
        {
            _cache = new Dictionary<string, ModelSection>();
        }

        public ModelSection GetConfig(string modelKey)
        {
            _rwLock.EnterReadLock();

            ModelSection config = null;

            try
            {
                config = _cache[modelKey];
            }
            finally
            {
                _rwLock.ExitReadLock();
            }

            return config;
        }

        public void SaveConfig(string modelKey, ModelSection config)
        {
            _rwLock.EnterWriteLock();

            try
            {
                _cache[modelKey] = config;
            }
            finally
            {
                _rwLock.ExitWriteLock();
            }
        }

        public bool ContainKey(string modelKey)
        {
            _rwLock.EnterReadLock();
            bool result = false;

            try
            {
                result = _cache.ContainsKey(modelKey);
            }
            finally
            {
                _rwLock.ExitReadLock();
            }

            return result;
        }
    }
}
